from __future__ import absolute_import, unicode_literals

import re
import sys
import types
import warnings

from contextlib import contextmanager
from functools import partial
from six import string_types, itervalues as values, iteritems as items

from . import mock

try:
    import unittest  # noqa
    unittest.skip
    from unittest.util import safe_repr, unorderable_list_difference
except AttributeError:
    import unittest2 as unittest  # noqa
    from unittest2.util import safe_repr, unorderable_list_difference  # noqa

__all__ = ['Case']


# -- adds assertWarns from recent unittest2, not in Python 2.7.

class _AssertRaisesBaseContext(object):

    def __init__(self, expected, test_case, callable_obj=None,
                 expected_regex=None):
        self.expected = expected
        self.failureException = test_case.failureException
        self.obj_name = None
        if isinstance(expected_regex, string_types):
            expected_regex = re.compile(expected_regex)
        self.expected_regex = expected_regex


def _is_magic_module(m):
    # some libraries create custom module types that are lazily
    # lodaded, e.g. Django installs some modules in sys.modules that
    # will load _tkinter and other shit when touched.

    # pyflakes refuses to accept 'noqa' for this isinstance.
    cls, modtype = type(m), types.ModuleType
    try:
        variables = vars(cls)
    except TypeError:
        return True
    else:
        return (cls is not modtype and (
            '__getattr__' in variables or
            '__getattribute__' in variables))


class _AssertWarnsContext(_AssertRaisesBaseContext):
    """A context manager used to implement TestCase.assertWarns* methods."""

    def __enter__(self):
        # The __warningregistry__'s need to be in a pristine state for tests
        # to work properly.
        warnings.resetwarnings()
        for v in list(values(sys.modules)):
            # do not evaluate Django moved modules and other lazily
            # initialized modules.
            if v and not _is_magic_module(v):
                # use raw __getattribute__ to protect even better from
                # lazily loaded modules
                try:
                    object.__getattribute__(v, '__warningregistry__')
                except AttributeError:
                    pass
                else:
                    object.__setattr__(v, '__warningregistry__', {})
        self.warnings_manager = warnings.catch_warnings(record=True)
        self.warnings = self.warnings_manager.__enter__()
        warnings.simplefilter('always', self.expected)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.warnings_manager.__exit__(exc_type, exc_value, tb)
        if exc_type is not None:
            # let unexpected exceptions pass through
            return
        try:
            exc_name = self.expected.__name__
        except AttributeError:
            exc_name = str(self.expected)
        first_matching = None
        for m in self.warnings:
            w = m.message
            if not isinstance(w, self.expected):
                continue
            if first_matching is None:
                first_matching = w
            if (self.expected_regex is not None and
                    not self.expected_regex.search(str(w))):
                continue
            # store warning for later retrieval
            self.warning = w
            self.filename = m.filename
            self.lineno = m.lineno
            return
        # Now we simply try to choose a helpful failure message
        if first_matching is not None:
            raise self.failureException(
                '%r does not match %r' % (
                    self.expected_regex.pattern, str(first_matching)))
        if self.obj_name:
            raise self.failureException(
                '%s not triggered by %s' % (exc_name, self.obj_name))
        else:
            raise self.failureException('%s not triggered' % exc_name)


class CaseMixin(object):
    """Mixin class that adds the utility methods to any unittest TestCase
    class."""

    def patch(self, *path, **options):
        """Patch object until test case returns.

        Example::

            from case import Case

            class test_Frobulator(Case):

                def setup(self):
                    frobulate = self.patch('some.where.Frobulator.frobulate')

        """
        manager = mock.patch('.'.join(path), **options)
        patched = manager.start()
        self.addCleanup(manager.stop)
        return patched

    def mock_modules(self, *mods):
        """Mock modules for the duration of the test.

        See :func:`case.mock.module`

        """
        modules = []
        for mod in mods:
            mod = mod.split('.')
            modules.extend(reversed([
                '.'.join(mod[:-i] if i else mod) for i in range(len(mod))
            ]))
        modules = sorted(set(modules))
        return self.wrap_context(mock.module(*modules))

    def mask_modules(self, *modules):
        """Make modules for the duration of the test.

        See :func:`case.mock.mask_modules`.

        """
        self.wrap_context(mock.mask_modules(*modules))

    def wrap_context(self, context):
        """Wrap context so that the context exits when the test completes."""
        ret = context.__enter__()
        self.addCleanup(partial(context.__exit__, None, None, None))
        return ret

    def mock_environ(self, env_name, env_value):
        """Mock environment variable value for the duration of the test.

        See :func:`case.mock.environ`.

        """
        return self.wrap_context(mock.environ(env_name, env_value))


class Case(unittest.TestCase, CaseMixin):
    """Test Case

    Subclass of :class:`unittest.TestCase` adding convenience
    methods.

    **setup / teardown**

    New :meth:`setup` and :meth:`teardown` methods can be defined
    in addition to the core :meth:`setUp` + :meth:`tearDown`
    methods.

    Note: If you redefine the core :meth:`setUp` + :meth:`tearDown`
          methods you must make sure ``super`` is called.
          ``super`` is not necessary for the lowercase versions.

    **Python 2.6 compatibility**

    This class also implements :meth:`assertWarns`, :meth:`assertWarnsRegex`,
    :meth:`assertDictContainsSubset`, and :meth:`assertItemsEqual`
    which are not available in the original Python 2.6 unittest
    implementation.

    """
    DeprecationWarning = DeprecationWarning
    PendingDeprecationWarning = PendingDeprecationWarning

    def setUp(self):
        self.setup()

    def tearDown(self):
        self.teardown()

    def setup(self):
        pass

    def teardown(self):
        pass

    def assertWarns(self, expected_warning):
        return _AssertWarnsContext(expected_warning, self, None)

    def assertWarnsRegex(self, expected_warning, expected_regex):
        return _AssertWarnsContext(expected_warning, self,
                                   None, expected_regex)

    @contextmanager
    def assertDeprecated(self):
        with self.assertWarnsRegex(self.DeprecationWarning,
                                   r'scheduled for removal'):
            yield

    @contextmanager
    def assertPendingDeprecation(self):
        with self.assertWarnsRegex(self.PendingDeprecationWarning,
                                   r'scheduled for deprecation'):
            yield

    def assertDictContainsSubset(self, expected, actual, msg=None):
        missing, mismatched = [], []

        for key, value in items(expected):
            if key not in actual:
                missing.append(key)
            elif value != actual[key]:
                mismatched.append('%s, expected: %s, actual: %s' % (
                    safe_repr(key), safe_repr(value),
                    safe_repr(actual[key])))

        if not (missing or mismatched):
            return

        standard_msg = ''
        if missing:
            standard_msg = 'Missing: %s' % ','.join(map(safe_repr, missing))

        if mismatched:
            if standard_msg:
                standard_msg += '; '
            standard_msg += 'Mismatched values: %s' % (
                ','.join(mismatched))

        self.fail(self._formatMessage(msg, standard_msg))

    def assertItemsEqual(self, expected_seq, actual_seq, msg=None):
        missing = unexpected = None
        try:
            expected = sorted(expected_seq)
            actual = sorted(actual_seq)
        except TypeError:
            # Unsortable items (example: set(), complex(), ...)
            expected = list(expected_seq)
            actual = list(actual_seq)
            missing, unexpected = unorderable_list_difference(
                expected, actual)
        else:
            return self.assertSequenceEqual(expected, actual, msg=msg)

        errors = []
        if missing:
            errors.append(
                'Expected, but missing:\n    %s' % (safe_repr(missing),)
            )
        if unexpected:
            errors.append(
                'Unexpected, but present:\n    %s' % (safe_repr(unexpected),)
            )
        if errors:
            standardMsg = '\n'.join(errors)
            self.fail(self._formatMessage(msg, standardMsg))



from __future__ import absolute_import, unicode_literals

import importlib
import os
import sys

from nose import SkipTest

from .utils import decorator, symbol_by_name

__all__ = [
    'todo',
    'if_darwin', 'unless_darwin',
    'if_environ', 'unless_environ',
    'if_module', 'unless_module',
    'if_jython', 'unless_jython',
    'if_platform', 'unless_platform',
    'if_pypy', 'unless_pypy',
    'if_python3', 'unless_python3',
    'if_win32', 'unless_win32',
    'if_symbol', 'unless_symbol',
    'if_python_version_after', 'if_python_version_before',
]


@decorator
def if_python_version_before(*version, **kwargs):
    """Skip test if Python version is less than ``*version``.

    Example::

        # skips test if running on Python < 3.1
        @skip.if_python_version_before(3, 1)

    """
    if sys.version_info < version:
        raise SkipTest('python < {0}: {1}'.format(
            '.'.join(map(str, version)),
            kwargs.get('reason') or 'incompatible'))
    yield


@decorator
def if_python_version_after(*version, **kwargs):
    """Skip test if Python version is greater or equal to ``*version``.

    Example::

        # skips test if running on Python >= 3.5
        @skip.if_python_version_after(3, 5)

    """
    if sys.version_info >= version:
        raise SkipTest('python >= {0}: {1}'.format(
            '.'.join(map(str, version)),
            kwargs.get('reason') or 'incompatible'))
    yield


def if_python3(*version, **kwargs):
    """Skip test if Python version is 3 or later.

    Example::

        @skip.if_python3(reason='does not have buffer type')

    """
    return if_python_version_after(3, *version, **kwargs)


def unless_python3(*version, **kwargs):
    """Skip test if Python version is Python 2 or earlier.

    Example::

        @skip.unless_python3()

    """
    return if_python_version_before(3, *version, **kwargs)


@decorator
def if_environ(env_var_name):
    """Skip test if environment variable ``env_var_name`` is defined.

    Example::

        @skip.if_environ('SKIP_SLOW_TESTS')

    """
    if os.environ.get(env_var_name):
        raise SkipTest('envvar {0} set'.format(env_var_name))
    yield


@decorator
def unless_environ(env_var_name):
    """Skip test if environment variable ``env_var_name`` is undefined.

    Example::

        @skip.unless_environ('LOCALE')

    """
    if not os.environ.get(env_var_name):
        raise SkipTest('envvar {0} not set'.format(env_var_name))
    yield


@decorator
def _skip_test(reason, sign):
    raise SkipTest('{0}: {1}'.format(sign, reason))


def todo(reason):
    """Skip test flagging case as TODO.

    Example::

        @skip.todo(reason='broken test')

    """
    return _skip_test(reason, sign='TODO')


def skip(reason):
    return _skip_test(reason, sign='SKIP')


@decorator
def if_module(module, name=None, import_errors=(ImportError,)):
    """Skip test if ``module`` can be imported.

    :param module: Module to import.
    :keyword name: Alternative module name to use in reason.
    :keyword import_errors: Tuple of import errors to check for.
        Default is ``(ImportError,)``.

    Example::

        @skip.if_module('librabbitmq')

    """
    try:
        importlib.import_module(module)
    except import_errors:
        pass
    else:
        raise SkipTest('module available: {0}'.format(name or module))
    yield


@decorator
def unless_module(module, name=None, import_errors=(ImportError,)):
    """Skip test if ``module`` can not be imported.

    :param module: Module to import.
    :keyword name: Alternative module name to use in reason.
    :keyword import_errors: Tuple of import errors to check for.
        Default is ``(ImportError,)``.

    Example::

        @skip.unless_module('librabbitmq')

    """
    try:
        importlib.import_module(module)
    except import_errors:
        raise SkipTest('module not installed: {0}'.format(name or module))
    yield


@decorator
def if_symbol(symbol, name=None,
              import_errors=(AttributeError, ImportError)):
    """Skip test if ``symbol`` can be imported.

    :param module: Symbol to import.
    :keyword name: Alternative symbol name to use in reason.
    :keyword import_errors: Tuple of import errors to check for.
        Default is ``(AttributeError, ImportError,)``.

    Example::

        @skip.if_symbol('django.db.transaction:on_commit')

    """
    try:
        symbol_by_name(symbol)
    except import_errors:
        pass
    else:
        raise SkipTest('symbol exists: {0}'.format(name or symbol))
    yield


@decorator
def unless_symbol(symbol, name=None,
                  import_errors=(AttributeError, ImportError)):
    """Skip test if ``symbol`` cannot be imported.

    :param module: Symbol to import.
    :keyword name: Alternative symbol name to use in reason.
    :keyword import_errors: Tuple of import errors to check for.
        Default is ``(AttributeError, ImportError,)``.

    Example::

        @skip.unless_symbol('django.db.transaction:on_commit')

    """
    try:
        symbol_by_name(symbol)
    except import_errors:
        raise SkipTest('missing symbol: {0}'.format(name or symbol))
    yield


@decorator
def if_platform(platform_name, name=None):
    """Skip test if :data:`sys.platform` name matches ``platform_name``.

    :param platform_name: Name to match with :data:`sys.platform`.
    :keyword name: Alternative name to use in reason.

    Example::

        @skip.if_platform('netbsd', name='NetBSD')

    """
    if sys.platform.startswith(platform_name):
        raise SkipTest('does not work on {0}'.format(platform_name or name))
    yield


def if_jython():
    """Skip test if running under Jython.

    Example::

        @skip.if_jython()

    """
    return if_platform('java', name='Jython')


def if_win32():
    """Skip test if running under Windows.

    Example::

        @skip.if_win32()

    """
    return if_platform('win32', name='Windows')


def if_darwin():
    """Skip test if running under OS X.

    Example::

        @skip.if_darwin()

    """
    return if_platform('darwin', name='OS X')


@decorator
def unless_platform(platform_name, name=None):
    """Skip test if :data:`sys.platform` name does not match ``platform_name``.

    :param platform_name: Name to match with :data:`sys.platform`.
    :keyword name: Alternative name to use in reason.

    Example::

        @skip.unless_platform('netbsd', name='NetBSD')

    """
    if not sys.platform.startswith(platform_name):
        raise SkipTest('only applicable on {0}'.format(platform_name or name))
    yield


def unless_jython():
    """Skip test if not running under Jython."""
    return unless_platform('java', name='Jython')


def unless_win32():
    """Skip test if not running under Windows."""
    return unless_platform('win32', name='Windows')


def unless_darwin():
    """Skip test if not running under OS X."""
    return unless_platform('darwin', name='OS X')


@decorator
def if_pypy(reason='does not work on PyPy'):
    """Skip test if running on PyPy.

    Example::

        @skip.if_pypy()

    """
    if getattr(sys, 'pypy_version_info', None):
        raise SkipTest(reason)
    yield


@decorator
def unless_pypy(reason='only applicable for PyPy'):
    """Skip test if not running on PyPy.

    Example::

        @skip.unless_pypy()

    """
    if not hasattr(sys, 'pypy_version_info'):
        raise SkipTest(reason)
    yield



from __future__ import absolute_import, unicode_literals

import pytest
import sys

from functools import wraps
from six import iteritems as items

from . import patch
from . import mock

sentinel = object()


class fixture_with_options(object):
    """Pytest fixture with options specified in separate decrorator.

    The decorated fixture MUST take the request fixture as first argument,
    but is free to use other fixtures.

    Example:
        @fixture_with_options()
        def sftp(request,
                 username='test_username',
                 password='test_password'):
            return {'username': username, 'password': password}

        @sftp.options(username='foo', password='bar')
        def test_foo(sftp):
            assert sftp['username'] == 'foo'
            assert sftp['password'] == 'bar'
    """

    def __init__(self, marker_name=None):
        self.marker_name = marker_name

    def __call__(self, fun):
        marker_name = self.marker_name or fun.__name__

        @pytest.fixture()
        @wraps(fun)
        def _inner(request, *args, **kwargs):
            marker = request.node.get_marker(marker_name)
            print(request.node)
            return fun(request, *args, **dict(marker.kwargs, **kwargs))

        def options(*args, **kwargs):
            return getattr(pytest.mark, marker_name)(*args, **kwargs)

        _inner.options = options
        _inner.__wrapped__ = fun
        return _inner


class _patching(object):

    def __init__(self, monkeypatch, request):
        self.monkeypatch = monkeypatch
        self.request = request

    def __getattr__(self, name):
        return getattr(self.monkeypatch, name)

    def __call__(self, path, value=sentinel, name=None,
                 new=mock.MagicMock, **kwargs):
        value = self._value_or_mock(value, new, name, path, **kwargs)
        self.monkeypatch.setattr(path, value)
        return value

    def object(self, target, attribute, *args, **kwargs):
        return _wrap_context(
            patch.object(target, attribute, *args, **kwargs),
            self.request)

    def _value_or_mock(self, value, new, name, path, **kwargs):
        if value is sentinel:
            value = new(name=name or path.rpartition('.')[2])
        for k, v in items(kwargs):
            setattr(value, k, v)
        return value

    def setattr(self, target, name=sentinel, value=sentinel, **kwargs):
        # alias to __call__ with the interface of pytest.monkeypatch.setattr
        if value is sentinel:
            value, name = name, None
        return self(target, value, name=name)

    def setitem(self, dic, name, value=sentinel, new=mock.MagicMock, **kwargs):
        # same as pytest.monkeypatch.setattr but default value is MagicMock
        value = self._value_or_mock(value, new, name, dic, **kwargs)
        self.monkeypatch.setitem(dic, name, value)
        return value

    def modules(self, *mods):
        modules = []
        for mod in mods:
            mod = mod.split('.')
            modules.extend(reversed([
                '.'.join(mod[:-i] if i else mod) for i in range(len(mod))
            ]))
        modules = sorted(set(modules))
        return _wrap_context(mock.module(*modules), self.request)


def _wrap_context(context, request):
    ret = context.__enter__()

    def fin():
        context.__exit__(*sys.exc_info())
    request.addfinalizer(fin)
    return ret


@pytest.fixture()
def patching(monkeypatch, request):
    """Monkeypath.setattr shortcut.

    Example:
        .. code-block:: python

        def test_foo(patching):
            # execv value here will be mock.MagicMock by default.
            execv = patching('os.execv')

            patching('sys.platform', 'darwin')  # set concrete value
            patching.setenv('DJANGO_SETTINGS_MODULE', 'x.settings')

            # val will be of type mock.MagicMock by default
            val = patching.setitem('path.to.dict', 'KEY')
    """
    return _patching(monkeypatch, request)


class _stdouts(object):

    def __init__(self, stdout, stderr):
        self.stdout = stdout
        self.stderr = stderr


@pytest.fixture()
def stdouts(request):
    return _stdouts(*_wrap_context(mock.stdouts(), request))



"""Python unittest Utilities"""
from __future__ import absolute_import, unicode_literals

import re

from collections import namedtuple

from .case import Case
from .mock import ANY, ContextMock, MagicMock, Mock, call, patch, sentinel

from . import mock
from . import skip

__version__ = '1.5.3'
__author__ = 'Ask Solem'
__contact__ = 'ask@celeryproject.org'
__homepage__ = 'http://github.com/celery/case'
__docformat__ = 'restructuredtext'

# -eof meta-

version_info_t = namedtuple('version_info_t', (
    'major', 'minor', 'micro', 'releaselevel', 'serial'
))

# bumpversion can only search for {current_version}
# so we have to parse the version here.
_temp = re.match(
    r'(\d+)\.(\d+).(\d+)(.+)?', __version__).groups()
VERSION = version_info = version_info_t(
    int(_temp[0]), int(_temp[1]), int(_temp[2]), _temp[3] or '', '')
del(_temp)
del(re)

__all__ = [
    b'Case',

    b'ANY', b'ContextMock', b'MagicMock', b'Mock',
    b'call', b'patch', b'sentinel',

    b'mock', b'skip',
]



from __future__ import absolute_import, unicode_literals

import importlib
import inspect
import logging
import os
import platform
import sys
import time
import types

from contextlib import contextmanager
from functools import wraps
from six import reraise, string_types, iteritems as items
from six.moves import builtins

from .utils import WhateverIO, decorator, get_logger_handlers, noop

try:
    from importlib import reload
except ImportError:
    try:
        from imp import reload
    except ImportError:
        reload = reload

try:
    from unittest import mock
except ImportError:
    import mock  # noqa

PY3 = sys.version_info[0] >= 3
if PY3:
    open_fqdn = 'builtins.open'
    module_name_t = str
else:
    open_fqdn = '__builtin__.open'  # noqa
    module_name_t = bytes  # noqa

__all__ = [
    'ANY', 'ContextMock', 'MagicMock', 'Mock', 'MockCallbacks',
    'call', 'patch', 'sentinel',

    'wrap_logger', 'environ', 'sleepdeprived', 'mask_modules', 'mute',
    'stdouts', 'replace_module_value', 'sys_version', 'pypy_version',
    'platform_pyimp', 'sys_platform', 'reset_modules', 'module',
    'open', 'restore_logging', 'module_exists', 'create_patcher',
]

ANY = mock.ANY
call = mock.call
sentinel = mock.sentinel


def create_patcher(*partial_path):

    def patcher(name, *args, **kwargs):
        return patch(".".join(partial_path + (name, )), *args, **kwargs)
    return patcher


class MockMixin(object):

    def on_nth_call_do(self, side_effect, n=1):
        """Change Mock side effect after ``n`` calls.

        Example::

            mock.on_nth_call_do(RuntimeError, n=3)

        """
        def on_call(*args, **kwargs):
            if self.call_count >= n:
                self.side_effect = side_effect
            return self.return_value
        self.side_effect = on_call
        return self

    def on_nth_call_do_raise(self, excA, excB, n=1):
        """Change exception raised after ``n`` calls.

        Mock will raise excA until called `n` times, which after
        it will raise excB'.

        Example::

            >>> mock.on_nth_call_do_raise(KeyError(), RuntimError(), n=3)
            >>> mock()
            KeyError()
            >>> mock()
            KeyError()
            >>> mock()
            KeyError()
            >>> mock()
            RuntimeError()

        """
        def on_call(*args, **kwargs):
            if self.call_count >= n:
                self.side_effect = excB
            raise excA
        self.side_effect = on_call
        return self

    def on_nth_call_return(self, retval, n=1):
        """Change Mock to return specific return value after ``n`` calls.

        Example::

            mock.on_nth_call_return('STOP', n=3)

        """

        def on_call(*args, **kwargs):
            if self.call_count >= n:
                self.return_value = retval
            return self.return_value
        self.side_effect = on_call
        return self

    def _mock_update_attributes(self, attrs={}, **kwargs):
        for key, value in items(attrs):
            setattr(self, key, value)

    def assert_not_called(_mock_self):  # noqa
        """assert that the mock was never called."""
        self = _mock_self
        if self.call_count != 0:
            msg = ("Expected '%s' to not have been called. Called %s times." %
                   (self._mock_name or 'mock', self.call_count))
            raise AssertionError(msg)

    def assert_called(_mock_self):  # noqa
        """assert that the mock was called at least once."""
        self = _mock_self
        if self.call_count == 0:
            msg = ("Expected '%s' to have been called." %
                   self._mock_name or 'mock')
            raise AssertionError(msg)

    def assert_called_once(_mock_self):  # noqa
        """assert that the mock was called only once."""
        self = _mock_self
        if not self.call_count == 1:
            msg = ("Expected '%s' to have been called once. Called %s times." %
                   (self._mock_name or 'mock', self.call_count))
            raise AssertionError(msg)


class Mock(mock.Mock, MockMixin):

    def __init__(self, *args, **kwargs):
        super(Mock, self).__init__(*args, **kwargs)
        self._mock_update_attributes(**kwargs)


class MagicMock(mock.MagicMock, MockMixin):

    def __init__(self, *args, **kwargs):
        super(MagicMock, self).__init__(*args, **kwargs)
        self._mock_update_attributes(**kwargs)


class _ContextMock(Mock):
    """Dummy class implementing __enter__ and __exit__
    as the :keyword:`with` statement requires these to be implemented
    in the class, not just the instance."""

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        pass


def ContextMock(*args, **kwargs):
    """Mock that mocks :keyword:`with` statement contexts."""
    obj = _ContextMock(*args, **kwargs)
    obj.attach_mock(_ContextMock(), '__enter__')
    obj.attach_mock(_ContextMock(), '__exit__')
    obj.__enter__.return_value = obj
    # if __exit__ return a value the exception is ignored,
    # so it must return None here.
    obj.__exit__.return_value = None
    return obj


def _patch_sig1(target,
                new=None, spec=None, create=None,
                spec_set=None, autospec=None, new_callable=None, **kwargs):
    # signature for mock.patch,
    # used to inject new `new_callable` argument default.
    return new, autospec, new_callable


def _patch_sig2(target, attribute,
                new=None, spec=None, create=False, spec_set=None,
                autospec=None, new_callable=None, **kwargs):
    # signature for mock.patch.multiple + mock.patch.object,
    # used to inject new `new_callable` argument default.
    return new, autospec, new_callable


def _patch_sig_multiple(target,
                        spec=None, create=False, spec_set=None,
                        autospec=None, new_callable=None, **kwargs):
    # New is always set, don't have to specify new_callable
    return True, autospec, new_callable


def _create_patcher(fun, signature):

    @wraps(fun)
    def patcher(*args, **kwargs):
        new, autospec, new_callable = signature(*args, **kwargs)
        if new is None and autospec is None and new_callable is None:
            kwargs.setdefault('new_callable', MagicMock)
        return fun(*args, **kwargs)

    return patcher


patch = _create_patcher(mock.patch, _patch_sig1)
patch.dict = mock.patch.dict
patch.multiple = _create_patcher(mock.patch.multiple, _patch_sig_multiple)
patch.object = _create_patcher(mock.patch.object, _patch_sig2)
patch.stopall = mock.patch.stopall
patch.TEST_PREFIX = mock.patch.TEST_PREFIX


def _bind(f, o):
    @wraps(f)
    def bound_meth(*fargs, **fkwargs):
        return f(o, *fargs, **fkwargs)
    return bound_meth


if PY3:  # pragma: no cover
    def _get_class_fun(meth):
        return meth

    def module_name(s):
        if isinstance(s, bytes):
            return s.decode()
        return s
else:
    def _get_class_fun(meth):  # noqa
        return meth.__func__

    def module_name(s):  # noqa
        if isinstance(s, unicode):
            return s.encode()
        return s


class MockCallbacks(object):

    def __new__(cls, *args, **kwargs):
        r = Mock(name=cls.__name__)
        _get_class_fun(cls.__init__)(r, *args, **kwargs)
        for key, value in items(vars(cls)):
            if key not in ('__dict__', '__weakref__', '__new__', '__init__'):
                if inspect.ismethod(value) or inspect.isfunction(value):
                    r.__getattr__(key).side_effect = _bind(value, r)
                else:
                    r.__setattr__(key, value)
        return r


@decorator
def wrap_logger(logger, loglevel=logging.ERROR):
    """Wrap :class:`logging.Logger` with a StringIO() handler.

    yields a StringIO handle.

    Example::

        with mock.wrap_logger(logger, loglevel=logging.DEBUG) as sio:
            ...
            sio.getvalue()

    """
    old_handlers = get_logger_handlers(logger)
    sio = WhateverIO()
    siohandler = logging.StreamHandler(sio)
    logger.handlers = [siohandler]

    try:
        yield sio
    finally:
        logger.handlers = old_handlers


@decorator
def environ(env_name, env_value):
    """Mock environment variable value.

    Example::

        @mock.environ('DJANGO_SETTINGS_MODULE', 'proj.settings')
        def test_other_settings(self):
            ...

    """
    sentinel = object()
    prev_val = os.environ.get(env_name, sentinel)
    os.environ[env_name] = env_value
    try:
        yield
    finally:
        if prev_val is sentinel:
            os.environ.pop(env_name, None)
        else:
            os.environ[env_name] = prev_val


@decorator
def sleepdeprived(module=time):
    """Mock time.sleep to do nothing.

    Example::

        @mock.sleepdeprived()  # < patches time.sleep
        @mock.sleepdeprived(celery.result)  # < patches celery.result.sleep

    """
    old_sleep, module.sleep = module.sleep, noop
    try:
        yield
    finally:
        module.sleep = old_sleep


# Taken from
# http://bitbucket.org/runeh/snippets/src/tip/missing_modules.py
@decorator
def mask_modules(*modnames):
    """Ban some modules from being importable inside the context

    For example::

        >>> with mask_modules('sys'):
        ...     try:
        ...         import sys
        ...     except ImportError:
        ...         print('sys not found')
        sys not found

        >>> import sys  # noqa
        >>> sys.version
        (2, 5, 2, 'final', 0)

    Or as a decorator::

        @mask_modules('sys')
        def test_foo(self):
            ...

    """
    realimport = builtins.__import__

    def myimp(name, *args, **kwargs):
        if name in modnames:
            raise ImportError('No module named %s' % name)
        else:
            return realimport(name, *args, **kwargs)

    builtins.__import__ = myimp
    try:
        yield
    finally:
        builtins.__import__ = realimport


@decorator
def stdouts():
    """Override `sys.stdout` and `sys.stderr` with `StringIO`
    instances.

    Decorator example::

        @mock.stdouts
        def test_foo(self, stdout, stderr):
            something()
            self.assertIn('foo', stdout.getvalue())

    Context example::

        with mock.stdouts() as (stdout, stderr):
            something()
            self.assertIn('foo', stdout.getvalue())

    """
    prev_out, prev_err = sys.stdout, sys.stderr
    prev_rout, prev_rerr = sys.__stdout__, sys.__stderr__
    mystdout, mystderr = WhateverIO(), WhateverIO()
    sys.stdout = sys.__stdout__ = mystdout
    sys.stderr = sys.__stderr__ = mystderr

    try:
        yield mystdout, mystderr
    finally:
        sys.stdout = prev_out
        sys.stderr = prev_err
        sys.__stdout__ = prev_rout
        sys.__stderr__ = prev_rerr


@decorator
def mute():
    """Redirect `sys.stdout` and `sys.stderr` to /dev/null, silencent them.
    Decorator example::
        @mock.mute
        def test_foo(self):
            something()
    Context example::
        with mock.mute():
            something()
    """
    prev_out, prev_err = sys.stdout, sys.stderr
    prev_rout, prev_rerr = sys.__stdout__, sys.__stderr__
    devnull = open(os.devnull, 'w')
    mystdout, mystderr = devnull, devnull
    sys.stdout = sys.__stdout__ = mystdout
    sys.stderr = sys.__stderr__ = mystderr

    try:
        yield
    finally:
        sys.stdout = prev_out
        sys.stderr = prev_err
        sys.__stdout__ = prev_rout
        sys.__stderr__ = prev_rerr


@decorator
def replace_module_value(module, name, value=None):
    """Mock module value, given a module, attribute name and value.

    Decorator example::

        @mock.replace_module_value(module, 'CONSTANT', 3.03)
        def test_foo(self):
            ...

    Context example::

        with mock.replace_module_value(module, 'CONSTANT', 3.03):
            ...

    """
    has_prev = hasattr(module, name)
    prev = getattr(module, name, None)
    if value:
        setattr(module, name, value)
    else:
        try:
            delattr(module, name)
        except AttributeError:
            pass
    try:
        yield
    finally:
        if prev is not None:
            setattr(module, name, prev)
        if not has_prev:
            try:
                delattr(module, name)
            except AttributeError:
                pass


def sys_version(value=None):
    """Mock :data:`sys.version_info`

    Decorator example::

        @mock.sys_version((3, 6, 1))
        def test_foo(self):
            ...

    Context example::

        with mock.sys_version((3, 6, 1)):
            ...

    """
    return replace_module_value(sys, 'version_info', value)


def pypy_version(value=None):
    """Mock :data:`sys.pypy_version_info`

    Decorator example::

        @mock.pypy_version((3, 6, 1))
        def test_foo(self):
            ...

    Context example::

        with mock.pypy_version((3, 6, 1)):
            ...

    """
    return replace_module_value(sys, 'pypy_version_info', value)


def platform_pyimp(value=None):
    """Mock :data:`platform.python_implementation`

    Decorator example::

        @mock.platform_pyimp('PyPy')
        def test_foo(self):
            ...

    Context example::

        with mock.platform_pyimp('PyPy'):
            ...

    """
    return replace_module_value(platform, 'python_implementation', value)


@decorator
def sys_platform(value=None):
    """Mock :data:`sys.platform`

    Decorator example::

        @mock.sys_platform('darwin')
        def test_foo(self):
            ...

    Context example::

        with mock.sys_platform('darwin'):
            ...

    """

    prev, sys.platform = sys.platform, value
    try:
        yield
    finally:
        sys.platform = prev


@decorator
def reset_modules(*modules):
    """Remove modules from :data:`sys.modules` by name,
    and reset back again when the test/context returns.

    Decorator example::

        @mock.reset_modules('celery.result', 'celery.app.base')
        def test_foo(self):
            pass

    Context example::

        with mock.reset_modules('celery.result', 'celery.app.base'):
            pass

    """
    prev = dict((k, sys.modules.pop(k))
                for k in modules if k in sys.modules)
    try:
        for k in modules:
            reload(importlib.import_module(k))
        yield
    finally:
        sys.modules.update(prev)


@decorator
def module(*names):
    """Mock one or modules such that every attribute is a :class:`Mock`."""
    prev = {}

    class MockModule(types.ModuleType):

        def __getattr__(self, attr):
            setattr(self, attr, Mock())
            return types.ModuleType.__getattribute__(self, attr)

    mods = []
    for name in names:
        try:
            prev[name] = sys.modules[name]
        except KeyError:
            pass
        mod = sys.modules[name] = MockModule(module_name(name))
        mods.append(mod)
    try:
        yield mods
    finally:
        for name in names:
            try:
                sys.modules[name] = prev[name]
            except KeyError:
                try:
                    del(sys.modules[name])
                except KeyError:
                    pass


@contextmanager
def mock_context(mock, typ=Mock):
    context = mock.return_value = Mock()
    context.__enter__ = typ()
    context.__exit__ = typ()

    def on_exit(*x):
        if x[0]:
            reraise(x[0], x[1], x[2])
    context.__exit__.side_effect = on_exit
    context.__enter__.return_value = context
    try:
        yield context
    finally:
        context.reset()


@decorator
def open(typ=WhateverIO, side_effect=None):
    """Patch builtins.open so that it returns StringIO object.

    :param typ: File object for open to return.
        Defaults to :class:`WhateverIO` which is the bastard child
        of :class:`io.StringIO` and :class:`io.BytesIO` accepting
        both bytes and unicode input.
    :param side_effect: Additional side effect for when the open context
        is entered.

    Decorator example::

        @mock.open()
        def test_foo(self, open_fh):
            something_opening_and_writing_a_file()
            self.assertIn('foo', open_fh.getvalue())

    Context example::

        with mock.open(io.BytesIO) as open_fh:
            something_opening_and_writing_bytes_to_a_file()
            self.assertIn(b'foo', open_fh.getvalue())

    """
    with patch(open_fqdn) as open_:
        with mock_context(open_) as context:
            if side_effect is not None:
                context.__enter__.side_effect = side_effect
            val = context.__enter__.return_value = typ()
            val.__exit__ = Mock()
            yield val


@decorator
def restore_logging():
    """Restore root logger handlers after test returns.

    Decorator example::

        @mock.restore_logging()
        def test_foo(self):
            setup_logging()

    Context example::

        with mock.restore_logging():
            setup_logging()

    """
    outs = sys.stdout, sys.stderr, sys.__stdout__, sys.__stderr__
    root = logging.getLogger()
    level = root.level
    handlers = root.handlers

    try:
        yield
    finally:
        sys.stdout, sys.stderr, sys.__stdout__, sys.__stderr__ = outs
        root.level = level
        root.handlers[:] = handlers


@decorator
def module_exists(*modules):
    """Patch one or more modules to ensure they exist.

    A module name with multiple paths (e.g. gevent.monkey) will
    ensure all parent modules are also patched (``gevent`` +
    ``gevent.monkey``).

    Decorator example::

        @mock.module_exists('gevent.monkey')
        def test_foo(self):
            pass

    Context example::

        with mock.module_exists('gevent.monkey'):
            gevent.monkey.patch_all = Mock(name='patch_all')
            ...

    """
    gen = []
    old_modules = []
    for module in modules:
        if isinstance(module, string_types):
            module = types.ModuleType(module_name(module))
        gen.append(module)
        if module.__name__ in sys.modules:
            old_modules.append(sys.modules[module.__name__])
        sys.modules[module.__name__] = module
        name = module.__name__
        if '.' in name:
            parent, _, attr = name.rpartition('.')
            setattr(sys.modules[parent], attr, module)
    try:
        yield
    finally:
        for module in gen:
            sys.modules.pop(module.__name__, None)
        for module in old_modules:
            sys.modules[module.__name__] = module



from __future__ import absolute_import, unicode_literals

import functools
import importlib
import inspect
import io
import logging
import sys
import unittest

from contextlib import contextmanager
from six import reraise, string_types

__all__ = [
    'WhateverIO', 'decorator', 'get_logger_handlers',
    'noop', 'symbol_by_name',
]

StringIO = io.StringIO
_SIO_write = StringIO.write
_SIO_init = StringIO.__init__


def update_wrapper(wrapper, wrapped, *args, **kwargs):
    wrapper = functools.update_wrapper(wrapper, wrapped, *args, **kwargs)
    wrapper.__wrapped__ = wrapped
    return wrapper


def wraps(wrapped,
          assigned=functools.WRAPPER_ASSIGNMENTS,
          updated=functools.WRAPPER_UPDATES):
    return functools.partial(update_wrapper, wrapped=wrapped,
                             assigned=assigned, updated=updated)


class _CallableContext(object):

    def __init__(self, context, cargs, ckwargs, fun):
        self.context = context
        self.cargs = cargs
        self.ckwargs = ckwargs
        self.fun = fun

    def __call__(self, *args, **kwargs):
        return self.fun(*args, **kwargs)

    def __enter__(self):
        self.ctx = self.context(*self.cargs, **self.ckwargs)
        return self.ctx.__enter__()

    def __exit__(self, *einfo):
        if self.ctx:
            return self.ctx.__exit__(*einfo)


def is_unittest_testcase(cls):
    try:
        cls.mro
    except AttributeError:
        pass  # py.test uses old style classes
    else:
        return issubclass(cls, unittest.TestCase)


def augment_setup(orig_setup, context, pargs, pkwargs):
    def around_setup_method(self, *args, **kwargs):
        try:
            contexts = self.__rb3dc_contexts__
        except AttributeError:
            contexts = self.__rb3dc_contexts = []
        p = context(*pargs, **pkwargs)
        p.__enter__()
        contexts.append(p)
        if orig_setup:
            return orig_setup(self, *args, **kwargs)
    if orig_setup:
        around_setup_method = wraps(orig_setup)(around_setup_method)
        around_setup_method.__wrapped__ = orig_setup
    return around_setup_method


def augment_teardown(orig_teardown, context, pargs, pkwargs):
    def around_teardown(self, *args, **kwargs):
        try:
            contexts = self.__rb3dc_contexts__
        except AttributeError:
            pass
        else:
            for context in contexts:
                context.__exit__(*sys.exc_info())
        if orig_teardown:
            orig_teardown(self, *args, **kwargs)
    if orig_teardown:
        around_teardown = wraps(orig_teardown)(around_teardown)
        around_teardown.__wrapped__ = orig_teardown
    return around_teardown


def decorator(predicate):
    context = contextmanager(predicate)

    @wraps(predicate)
    def take_arguments(*pargs, **pkwargs):

        @wraps(predicate)
        def decorator(cls):
            if inspect.isclass(cls):
                if is_unittest_testcase(cls):
                    orig_setup = cls.setUp
                    orig_teardown = cls.tearDown
                    cls.setUp = augment_setup(
                        orig_setup, context, pargs, pkwargs)
                    cls.tearDown = augment_teardown(
                        orig_teardown, context, pargs, pkwargs)
                else:  # py.test
                    orig_setup = getattr(cls, 'setup_method', None)
                    orig_teardown = getattr(cls, 'teardown_method', None)
                    cls.setup_method = augment_setup(
                        orig_setup, context, pargs, pkwargs)
                    cls.teardown_method = augment_teardown(
                        orig_teardown, context, pargs, pkwargs)
                return cls
            else:
                @wraps(cls)
                def around_case(*args, **kwargs):
                    with context(*pargs, **pkwargs) as context_args:
                        context_args = context_args or ()
                        if not isinstance(context_args, tuple):
                            context_args = (context_args,)
                        return cls(*args + context_args, **kwargs)
                return around_case

        if len(pargs) == 1 and callable(pargs[0]):
            fun, pargs = pargs[0], ()
            return decorator(fun)
        return _CallableContext(context, pargs, pkwargs, decorator)
    assert take_arguments.__wrapped__
    return take_arguments


def get_logger_handlers(logger):
    return [
        h for h in logger.handlers
        if not isinstance(h, logging.NullHandler)
    ]


def symbol_by_name(name, aliases={}, imp=None, package=None,
                   sep='.', default=None, **kwargs):
    """Get symbol by qualified name.

    The name should be the full dot-separated path to the class::

        modulename.ClassName

    Example::

        celery.concurrency.processes.TaskPool
                                    ^- class name

    or using ':' to separate module and symbol::

        celery.concurrency.processes:TaskPool

    If `aliases` is provided, a dict containing short name/long name
    mappings, the name is looked up in the aliases first.

    Examples:

        >>> symbol_by_name('celery.concurrency.processes.TaskPool')
        <class 'celery.concurrency.processes.TaskPool'>

        >>> symbol_by_name('default', {
        ...     'default': 'celery.concurrency.processes.TaskPool'})
        <class 'celery.concurrency.processes.TaskPool'>

        # Does not try to look up non-string names.
        >>> from celery.concurrency.processes import TaskPool
        >>> symbol_by_name(TaskPool) is TaskPool
        True

    """
    if imp is None:
        imp = importlib.import_module

    if not isinstance(name, string_types):
        return name                                 # already a class

    name = aliases.get(name) or name
    sep = ':' if ':' in name else sep
    module_name, _, cls_name = name.rpartition(sep)
    if not module_name:
        cls_name, module_name = None, package if package else cls_name
    try:
        try:
            module = imp(module_name, package=package, **kwargs)
        except ValueError as exc:
            reraise(ValueError,
                    ValueError("Couldn't import {0!r}: {1}".format(name, exc)),
                    sys.exc_info()[2])
        return getattr(module, cls_name) if cls_name else module
    except (ImportError, AttributeError):
        if default is None:
            raise
    return default


class WhateverIO(StringIO):

    def __init__(self, v=None, *a, **kw):
        _SIO_init(self, v.decode() if isinstance(v, bytes) else v, *a, **kw)

    def write(self, data):
        _SIO_write(self, data.decode() if isinstance(data, bytes) else data)


def noop(*args, **kwargs):
    pass
